@@ -16,6 +16,7 @@ import asyncio
 
 from .constants import EXTENDED_DATA_STDERR
 from .misc import BreakReceived, SignalReceived, TerminalSizeChanged
+from .misc import async_iterator, python35
 from .session import SSHClientSession, SSHServerSession
 from .session import SSHTCPSession, SSHUNIXSession
 from .sftp import SFTPServerHandler
@@ -29,6 +30,24 @@ class SSHReader:
         self._chan = chan
         self._datatype = datatype
 
+    if python35:
+        @async_iterator
+        def __aiter__(self):
+            """Allow SSHReader to be an async iterator"""
+
+            return self
+
+        @asyncio.coroutine
+        def __anext__(self):
+            """Return one line at a time when used as an async iterator"""
+
+            line = yield from self.readline()
+
+            if line:
+                return line
+            else:
+                raise StopAsyncIteration
+
     @property
     def channel(self):
         """The SSH channel associated with this stream"""
@@ -47,6 +66,7 @@ class SSHReader:
 
         return self._chan.get_extra_info(name, default)
 
+    @asyncio.coroutine
     def read(self, n=-1):
         """Read data from the stream
 
@@ -67,6 +87,7 @@ class SSHReader:
 
         return self._session.read(n, self._datatype, exact=False)
 
+    @asyncio.coroutine
     def readline(self):
         """Read one line from the stream
 
@@ -77,10 +98,15 @@ class SSHReader:
            line is returned. If EOF was received and the receive buffer
            is empty, an empty bytes or str object is returned.
 
+           .. note:: In Python 3.5 and later, :class:`SSHReader` objects
+                     can also be used as async iterators, returning input
+                     data one line at a time.
+
         """
 
         return self._session.readline(self._datatype)
 
+    @asyncio.coroutine
     def readexactly(self, n):
         """Read an exact amount of data from the stream
 
@@ -106,6 +132,11 @@ class SSHReader:
 
         return self._session.at_eof(self._datatype)
 
+    def get_redirect_info(self):
+        """Get information needed to redirect from this SSHReader"""
+
+        return self._session, self._datatype
+
 
 class SSHWriter:
     """SSH write stream handler"""
@@ -160,7 +191,7 @@ class SSHWriter:
 
         """
 
-        return (yield from self._session.drain())
+        return (yield from self._session.drain(self._datatype))
 
     def write(self, data):
         """Write data to the stream
@@ -196,6 +227,11 @@ class SSHWriter:
 
         return self._chan.write_eof()
 
+    def get_redirect_info(self):
+        """Get information needed to redirect to this SSHWriter"""
+
+        return self._session, self._datatype
+
 
 class SSHStreamSession:
     """SSH stream session handler"""
@@ -203,6 +239,7 @@ class SSHStreamSession:
     def __init__(self):
         self._chan = None
         self._conn = None
+        self._encoding = None
         self._loop = None
         self._limit = None
         self._exception = None
@@ -238,6 +275,13 @@ class SSHStreamSession:
             if not waiter.done():
                 waiter.set_result(None)
 
+    def _should_block_drain(self, datatype):
+        """Return whether output is still being written to the channel"""
+
+        # pylint: disable=unused-argument
+
+        return self._write_paused
+
     def _unblock_drain(self):
         """Signal that more data can be written on the stream"""
 
@@ -245,11 +289,37 @@ class SSHStreamSession:
             if not waiter.done(): # pragma: no branch
                 waiter.set_result(None)
 
+    def _should_pause_reading(self):
+        """Return whether to pause reading from the channel"""
+
+        return self._limit and self._recv_buf_len >= self._limit
+
+    def _maybe_pause_reading(self):
+        """Pause reading if necessary"""
+
+        if not self._read_paused and self._should_pause_reading():
+            self._read_paused = True
+            self._chan.pause_reading()
+            return True
+        else:
+            return False
+
+    def _maybe_resume_reading(self):
+        """Resume reading if necessary"""
+
+        if self._read_paused and not self._should_pause_reading():
+            self._read_paused = False
+            self._chan.resume_reading()
+            return True
+        else:
+            return False
+
     def connection_made(self, chan):
         """Handle a newly opened channel"""
 
         self._chan = chan
         self._conn = chan.get_connection()
+        self._encoding = chan.get_encoding()
         self._loop = chan.get_loop()
         self._limit = self._chan.get_recv_window()
 
@@ -279,15 +349,13 @@ class SSHStreamSession:
         self._recv_buf[datatype].append(data)
         self._recv_buf_len += len(data)
         self._unblock_read(datatype)
-
-        if self._recv_buf_len >= self._limit:
-            self._read_paused = True
-            self._chan.pause_reading()
+        self._maybe_pause_reading()
 
     def eof_received(self):
         """Handle an incoming end of file on the channel"""
 
         self._eof_received = True
+
         for datatype in self._read_waiter.keys():
             self._unblock_read(datatype)
 
@@ -314,7 +382,7 @@ class SSHStreamSession:
         """Read data from the channel"""
 
         recv_buf = self._recv_buf[datatype]
-        buf = '' if self._chan.get_encoding() else b''
+        buf = '' if self._encoding else b''
         data = []
 
         while True:
@@ -337,9 +405,7 @@ class SSHStreamSession:
                 self._recv_buf_len -= l
                 n -= l
 
-            if self._read_paused and self._recv_buf_len < self._limit:
-                self._read_paused = False
-                self._chan.resume_reading()
+            if self._maybe_resume_reading():
                 continue
 
             if n == 0 or (n > 0 and data and not exact) or self._eof_received:
@@ -358,7 +424,7 @@ class SSHStreamSession:
         """Read a line from the channel"""
 
         recv_buf = self._recv_buf[datatype]
-        buf, sep = ('', '\n') if self._chan.get_encoding() else (b'', b'\n')
+        buf, sep = ('', '\n') if self._encoding else (b'', b'\n')
         data = []
 
         while True:
@@ -375,19 +441,17 @@ class SSHStreamSession:
                     recv_buf[0] = recv_buf[0][idx:]
                     self._recv_buf_len -= idx
 
-                    if self._read_paused and self._recv_buf_len < self._limit:
-                        self._read_paused = False
-                        self._chan.resume_reading()
+                    if not recv_buf[0]:
+                        recv_buf.pop(0)
 
+                    self._maybe_resume_reading()
                     return buf.join(data)
 
                 l = len(recv_buf[0])
                 data.append(recv_buf.pop(0))
                 self._recv_buf_len -= l
 
-            if self._read_paused:
-                self._read_paused = False
-                self._chan.resume_reading()
+            if self._maybe_resume_reading():
                 continue
 
             if self._eof_received:
@@ -396,10 +460,10 @@ class SSHStreamSession:
             yield from self._block_read(datatype)
 
     @asyncio.coroutine
-    def drain(self):
+    def drain(self, datatype):
         """Wait for data written to the channel to drain"""
 
-        if self._write_paused and not self._connection_lost:
+        while self._should_block_drain(datatype) and not self._connection_lost:
             try:
                 waiter = asyncio.Future(loop=self._loop)
                 self._drain_waiters.append(waiter)
@@ -453,6 +517,7 @@ class SSHServerStreamSession(SSHStreamSession, SSHServerSession):
 
         if self._chan.get_subsystem() == 'sftp':
             self._chan.set_encoding(None)
+            self._encoding = None
 
             handler = SFTPServerHandler(self._sftp_factory, self._conn,
                                         SSHReader(self, self._chan),
