@@ -55,10 +55,10 @@ class _AsyncPublicKeyClient(_PublicKeyClient):
 class _PublicKeyServer(Server):
     """Server for testing public key authentication"""
 
-    def __init__(self, set_keys=False):
+    def __init__(self, authorized_keys=None):
         super().__init__()
-        self._client_key = None
-        self._set_keys = set_keys
+        self._client_keys = []
+        self._authorized_keys = authorized_keys
 
     def connection_made(self, conn):
         """Called when a connection is made"""
@@ -69,10 +69,11 @@ class _PublicKeyServer(Server):
     def begin_auth(self, username):
         """Handle client authentication request"""
 
-        if self._set_keys:
-            self._conn.set_authorized_keys('authorized_keys')
+        if self._authorized_keys:
+            self._conn.set_authorized_keys(self._authorized_keys)
         else:
-            self._client_key = asyncssh.read_public_key('ckey.pub')
+            self._client_keys = asyncssh.load_public_keys(['ckey_dsa.pub',
+                                                           'ckey.pub'])
 
         return True
 
@@ -84,12 +85,12 @@ class _PublicKeyServer(Server):
     def validate_public_key(self, username, key):
         """Return whether key is an authorized client key for this user"""
 
-        return key == self._client_key
+        return key in self._client_keys
 
     def validate_ca_key(self, username, key):
         """Return whether key is an authorized CA key for this user"""
 
-        return key == self._client_key
+        return key in self._client_keys
 
 
 class _AsyncPublicKeyServer(_PublicKeyServer):
@@ -324,6 +325,17 @@ class _TestPublicKeyAuth(ServerTestCase):
 
         yield from conn.wait_closed()
 
+    @asynctest
+    def test_agent_signature_algs(self):
+        """Test ssh-agent keys with specific signature algorithms"""
+
+        for alg in ('ssh-rsa', 'rsa-sha2-256', 'rsa-sha2-512'):
+            with (yield from self.connect(username='ckey',
+                                          signature_algs=[alg])) as conn:
+                pass
+
+            yield from conn.wait_closed()
+
     @asynctest
     def test_agent_auth_failure(self):
         """Test failure connecting with ssh-agent authentication"""
@@ -352,6 +364,35 @@ class _TestPublicKeyAuth(ServerTestCase):
 
         yield from conn.wait_closed()
 
+    @asynctest
+    def test_public_key_signature_algs(self):
+        """Test public key authentication with specific signature algorithms"""
+
+        for alg in ('ssh-rsa', 'rsa-sha2-256', 'rsa-sha2-512'):
+            with (yield from self.connect(username='ckey', client_keys='ckey',
+                                          signature_algs=[alg])) as conn:
+                pass
+
+            yield from conn.wait_closed()
+
+    @asynctest
+    def test_no_server_signature_algs(self):
+        """Test a server which doesn't advertise signature algorithms"""
+
+        def skip_ext_info(self):
+            """Don't send extension information"""
+
+            # pylint: disable=unused-argument
+
+            return []
+
+        with patch('asyncssh.connection.SSHConnection._get_ext_info_kex_alg',
+                   skip_ext_info):
+            with (yield from self.connect(username='ckey')) as conn:
+                pass
+
+        yield from conn.wait_closed()
+
     @asynctest
     def test_default_public_key_auth(self):
         """Test connecting with default public key authentication"""
@@ -388,18 +429,30 @@ class _TestPublicKeyAuth(ServerTestCase):
         yield from conn.wait_closed()
 
     @asynctest
-    def test_client_key_sshkeypairs(self):
+    def test_client_key_keypairs(self):
         """Test client keys passed in as a list of SSHKeyPairs"""
 
-        agent = yield from asyncssh.connect_agent()
-        keylist = yield from agent.get_keys()
+        keys = asyncssh.load_keypairs('ckey')
 
         with (yield from self.connect(username='ckey',
-                                      client_keys=keylist)) as conn:
+                                      client_keys=keys)) as conn:
             pass
 
         yield from conn.wait_closed()
 
+    @asynctest
+    def test_client_key_agent_keypairs(self):
+        """Test client keys passed in as a list of SSHAgentKeyPairs"""
+
+        agent = yield from asyncssh.connect_agent()
+
+        for key in (yield from agent.get_keys()):
+            with (yield from self.connect(username='ckey',
+                                          client_keys=[key])) as conn:
+                pass
+
+        yield from conn.wait_closed()
+
         agent.close()
 
     @asynctest
@@ -532,6 +585,38 @@ class _TestPublicKeyAsyncServerAuth(_TestPublicKeyAuth):
         return (yield from cls.create_server(_AsyncPublicKeyServer))
 
 
+class _TestLimitedSignatureAlgs(ServerTestCase):
+    """Unit tests for limited public key signature algorithms"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which supports public key authentication"""
+
+        return (yield from cls.create_server(
+            _PublicKeyServer, authorized_client_keys='authorized_keys',
+            signature_algs=['ssh-rsa', 'rsa-sha2-512']))
+
+    @asynctest
+    def test_mismatched_signature_algs(self):
+        """Test mismatched signature algorithms"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(username='ckey', client_keys='ckey',
+                                    signature_algs=['rsa-sha2-256'])
+
+    @asynctest
+    def test_signature_alg_fallback(self):
+        """Test fall back to default signature algorithm"""
+
+        with (yield from self.connect(username='ckey', client_keys='ckey',
+                                      signature_algs=['rsa-sha2-256',
+                                                      'ssh-rsa'])) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+
 class _TestSetAuthorizedKeys(ServerTestCase):
     """Unit tests for public key authentication with set_authorized_keys"""
 
@@ -543,7 +628,7 @@ class _TestSetAuthorizedKeys(ServerTestCase):
         def server_factory():
             """Return an SSH server which calls set_authorized_keys"""
 
-            return _PublicKeyServer(set_keys=True)
+            return _PublicKeyServer(authorized_keys='authorized_keys')
 
         return (yield from cls.create_server(server_factory))
 
@@ -573,6 +658,33 @@ class _TestSetAuthorizedKeys(ServerTestCase):
         yield from conn.wait_closed()
 
 
+class _TestPreloadedAuthorizedKeys(ServerTestCase):
+    """Unit tests for authentication with pre-loaded authorized keys"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which supports public key authentication"""
+
+        def server_factory():
+            """Return an SSH server which calls set_authorized_keys"""
+
+            authorized_keys = asyncssh.read_authorized_keys('authorized_keys')
+            return _PublicKeyServer(authorized_keys=authorized_keys)
+
+        return (yield from cls.create_server(server_factory))
+
+    @asynctest
+    def test_pre_loaded_authorized_keys(self):
+        """Test set_authorized_keys with pre-loaded authorized keys"""
+
+        with (yield from self.connect(username='ckey',
+                                      client_keys='ckey')) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+
 class _TestPasswordAuth(ServerTestCase):
     """Unit tests for password authentication"""
 
